//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : Receptor
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "receptor.h"
#include <algorithm> // shimoda

#define NEWVOXEL_rPSC
#define NEWVOXEL_ELEC

//============================================================================//
void Receptor::rpscace(const float &aceratio, const int &num_grid, float *grid_coord,
                       float *mol_coord, int *iwork, float *fwork,
                       const float &param_rec_core, const float &param_lig_core)
//============================================================================//
{
    const float rho = -3.0;         // MEGADOCK parameter
    const float epsilon = param_rec_core;   // Tuning Parameter
    const float open_space = -77.0;    // temporary score for open space
    const float swollen_surface = -88.0; // temporary score for swollen receptor surface
    const float grid_width = 1.2;

    int       *nearesta;
    int       *surf_atom;
    int       *counter;
    float     *grid_r, *grid_i;
    float     *rjudge2, *radius_core2, *radius_surf2;
    float     *distance3;
    float     *xd, *yd, *zd;

    const int ng1 = num_grid;
    const int ng2 = ng1 * ng1;
    const int ng3 = ng1 * ng1 * ng1;
    const int na  = num_atoms();
    const int nag = num_atoms() * ng1;


    float *voxel_rpscace;
    voxel_rpscace = new float[ng3];


    for (int i=0; i<ng3; i++) {
    	if(voxel_rpscace[i]!=0.0){
    		voxel_rpscace[i]=0.0;
    	}
    	if(fwork[i]!=0.0){
    		fwork[i]=0.0;
    	}
    }

    nearesta  = iwork;
    surf_atom = &iwork[ng3];
    counter   = &iwork[ng3+na];
    grid_r    = fwork;
    grid_i    = &fwork[ng3];
    rjudge2   = &fwork[ng3*2];
    radius_core2  = &fwork[ng3*2+na];
    radius_surf2  = &fwork[ng3*2+na*2];
    distance3     = &fwork[ng3*2+na*3];
    xd        = &fwork[ng3*3+na*3];
    yd        = &fwork[ng3*3+na*3+nag];
    zd        = &fwork[ng3*3+na*3+nag*2];

    precalc(ng1,nearesta,rjudge2,radius_core2,radius_surf2);
    voxel_init(ng1,grid_coord,mol_coord,grid_r,grid_i,distance3,xd,yd,zd);


#ifdef NEWVOXEL_rPSC

// ##############################
// #fill up grid by atom-sphere #
// ##############################

    //*
    for( int l = 0 ; l < na ; l++ ) {// atom-grid assignment
        const int s_space = 2;
        const int lc = ng1 * l;
        const int i2 = mol_coord[3*l  ] / grid_width + ng1 / 2;
        const int j2 = mol_coord[3*l+1] / grid_width + ng1 / 2;
        const int k2 = mol_coord[3*l+2] / grid_width + ng1 / 2;
        const int ia = max(i2 - s_space, 0);
        const int ja = max(j2 - s_space, 0);
        const int ka = max(k2 - s_space, 0);
        const int ib = min(i2 + s_space+1, ng1);
        const int jb = min(j2 + s_space+1, ng1);
        const int kb = min(k2 + s_space+1, ng1);

        for( int i = ia ; i < ib ; i++ ) {// grid around atom[l]
            if(xd[lc+i] > radius_core2[l]) continue;

            for( int j = ja ; j < jb ; j++ ) {
                const float d2 = xd[lc+i]+yd[lc+j];
                if(d2 > radius_core2[l]) continue;

                for( int k = ka ; k < kb ; k++ ) {
                    const float d3 = xd[lc+i] + yd[lc+j] + zd[lc+k];
                    if( d3 < radius_core2[l] ) {// distance(grid-atom) < Van der Waals radius (* core)
                        grid_r[ng2*i+ng1*j+k] = epsilon;    // grid[i] is filled up by atom[l]
                    }
                }
            }
        }
    }
    //*/


// #####################
// #fill up cavity     #
// #####################

    for (int ij=0; ij<ng2; ij++) {// fill up cavity (phase1, 1/2)
        const int i = ij % ng1;
        const int j = ij / ng1;

        for (int k=0; k<ng1; k++) {// X, forward
            if(grid_r[i+ng1*j+ng2*k]!=epsilon) {
                grid_r[i+ng1*j+ng2*k]=open_space;
            } else {
                break;
            }
        }

        for (int k=ng1-1; k>=0; k--) {// X, backward
            if(grid_r[i+ng1*j+ng2*k]!=epsilon) {
                grid_r[i+ng1*j+ng2*k]=open_space;
            } else {
                break;
            }
        }

        for (int k=0; k<ng1; k++) {// Y, forward
            if(grid_r[k+ng1*i+ng2*j]!=epsilon) {
                grid_r[k+ng1*i+ng2*j]=open_space;
            } else {
                break;
            }
        }

        for (int k=ng1-1; k>=0; k--) {// Y, backward
            if(grid_r[k+ng1*i+ng2*j]!=epsilon) {
                grid_r[k+ng1*i+ng2*j]=open_space;
            } else {
                break;
            }
        }

        for (int k=0; k<ng1; k++) {// Z, forward
            if(grid_r[j+ng1*k+ng2*i]!=epsilon) {
                grid_r[j+ng1*k+ng2*i]=open_space;
            } else {
                break;
            }
        }

        for (int k=ng1-1; k>=0; k--) {// Z, backward
            if(grid_r[j+ng1*k+ng2*i]!=epsilon) {
                grid_r[j+ng1*k+ng2*i]=open_space;
            } else {
                break;
            }
        }
    }

    //*
    for (int ink=0; ink<2; ink++) {// fill up cavity (phase1, 2/2)

        for (int ijk=ng2; ijk<ng3-ng2; ijk++) {
            const int i = ijk / ng2;
            const int j = (ijk / ng1) % ng1;
            if(j<2||j>ng1-2) continue; // skip border
            const int k = ijk % ng1;
            if(k<2||k>ng1-2) continue; // skip border

            if(grid_r[ijk]==open_space) {
                if(grid_r[ijk+1]==0) {
                    grid_r[ijk+1]=open_space;
                }
                if(grid_r[ijk-1]==0) {
                    grid_r[ijk-1]=open_space;
                }
                if(grid_r[ijk+ng1]==0) {
                    grid_r[ijk+ng1]=open_space;
                }
                if(grid_r[ijk-ng1]==0) {
                    grid_r[ijk-ng1]=open_space;
                }
                if(grid_r[ijk+ng2]==0) {
                    grid_r[ijk+ng2]=open_space;
                }
                if(grid_r[ijk-ng2]==0) {
                    grid_r[ijk-ng2]=open_space;
                }
            }
        }
    }
    //*/

    for(int i=0; i<ng3; i++) {// fill up cavity (phase2)
        if(grid_r[i]==0) {
            grid_r[i]=epsilon;
        }
    }

    for(int i=0; i<ng3; i++) {// restore open_space to 0 (phase3)
        if(grid_r[i]==open_space) {
            grid_r[i]=0;
        }
    }

// #################################
// # set surface score of receptor #
// #################################

    for( int l = 0 ; l < na ; l++ ) {// atom-grid assignment
        const int s_space = 5;
        const int lc = ng1 * l;
        const int i2 = mol_coord[3*l  ] / grid_width + ng1 / 2;
        const int j2 = mol_coord[3*l+1] / grid_width + ng1 / 2;
        const int k2 = mol_coord[3*l+2] / grid_width + ng1 / 2;
        const int ia = max(i2 - s_space, 0);
        const int ja = max(j2 - s_space, 0);
        const int ka = max(k2 - s_space, 0);
        const int ib = min(i2 + s_space+1, ng1);
        const int jb = min(j2 + s_space+1, ng1);
        const int kb = min(k2 + s_space+1, ng1);

        for( int i = ia ; i < ib ; i++ ) {// grid around atom[l]
            if(xd[lc+i] > rjudge2[l]) continue;

            for( int j = ja ; j < jb ; j++ ) {
                const float d2 = xd[lc+i]+yd[lc+j];
                if(d2 > rjudge2[l]) continue;

                for( int k = ka ; k < kb ; k++ ) {
                    const int ijk = ng2*i+ng1*j+k;
                    const float d3 = xd[lc+i] + yd[lc+j] + zd[lc+k];

                    if( d3 < rjudge2[l]) {// distance(grid-atom) < Van der Waals radius + cutoff D(=3.6ang)
                        //counter[ijk] ++;    // grid[i] is filled up by atom[l]
                        voxel_rpscace[ijk] += 1.0 + _ACE[l] * (-0.8) * aceratio;
                        // rPSC value(1.0) + ACE value (0.8 is weight parameter) * w_h
                    }
                }
            }
        }
    }

    for( int ijk = 0 ; ijk < ng3 ; ijk++ ) {
        if( grid_r[ijk]==epsilon ) continue;

        grid_r[ijk] = voxel_rpscace[ijk];
    }

#else //NEWVOXEL ####################################################
// megadock 2.5.3 voxel

    for( int i = 0 ; i < ng3 ; i++ ) {
        counter[i] = 0;
        voxel_rpscace[i] = 0.0;
    }

    int   i, j, k, lc, kb;
    float d2, d3;

    #pragma omp parallel for private(i,j,k,lc,d2,d3,kb)
    for( int l = 0 ; l < na ; l++ ) {
        lc = ng1 * l;

        for( i = 0 ; i < ng1 ; i++ ) {
            if( xd[lc+i] > rjudge2[l] ) continue;

            for( j = 0 ; j < ng1 ; j++ ) {
                if( yd[lc+j] > rjudge2[l] ) continue;
                d2 = xd[lc+i] + yd[lc+j];

                for( k = 0 ; k < ng1 ; k++ ) {
                    d3 = d2 + zd[lc+k];
                    if( d3 > rjudge2[l] ) continue;
                    kb = ng2*i + ng1*j + k;

                    #pragma omp critical
                    {
                        if( d3 < distance3[kb] ) {
                            distance3[kb] = d3;
                            nearesta[kb] = l;
                        }
                        counter[kb] ++;
                        voxel_rpscace[kb] += 1.0 + _ACE[l] * (-0.8) * aceratio; // rPSC value(1.0) + ACE value (0.8 is weight parameter) * w_h
                    }
                }
            }
        }
    }

    for( int i = 0 ; i < ng3 ; i++ ) {
        if( nearesta[i] == -99 ) continue;

        if( distance3[i] <= rjudge2[nearesta[i]] ) {
            if( distance3[i] <= radius_core2[nearesta[i]] ) {
                grid_r[i] = 9.0 * rho;
            } else {
                grid_r[i] = voxel_rpscace[i];
            }
        }
    }

    int   nsurf = 0;

    for( int i = 1 ; i < ng1-1 ; i++ ) {
        const int ib = ng2 * i;

        for( int j = 1 ; j < ng1-1 ; j++ ) {
            const int jb = ib + ng1 * j;

            for( int k = 1 ; k < ng1-1 ; k++ ) {
                const int kb = jb + k;

                if( grid_r[kb] == 9.0*rho ) {
                    /*
                    if( grid_r[kb-ng2] == (float)counter[kb-ng2] ||
                        grid_r[kb+ng2] == (float)counter[kb+ng2] ||
                        grid_r[kb-ng1] == (float)counter[kb-ng1] ||
                        grid_r[kb+ng1] == (float)counter[kb+ng1] ||
                        grid_r[kb-1  ] == (float)counter[kb-1  ] ||
                        grid_r[kb+1  ] == (float)counter[kb+1  ] ) {
                    */
// Bug start
                    if( grid_r[kb-ng2] == 0.0 ||
                            grid_r[kb+ng2] == 0.0 ||
                            grid_r[kb-ng1] == 0.0 ||
                            grid_r[kb+ng1] == 0.0 ||
                            grid_r[kb-1  ] == 0.0 ||
                            grid_r[kb+1  ] == 0.0 ) {
// Bug end
                        if( distance3[kb] > radius_surf2[nearesta[kb]] ) {
                            grid_r[kb] = -9999.0;
                        } else {
                            int l;

                            for( l = 0 ; l < nsurf ; l++ ) {
                                if( surf_atom[l] == nearesta[kb] ) {
                                    break;
                                }
                            }

                            if( l >= nsurf ) {
                                surf_atom[nsurf++] = nearesta[kb];
                            }
                        }
                    }
                }
            }
        }
    }

    for( int i = ng2 ; i < ng3-ng2 ; i++ ) {
        if( grid_r[i] == -9999.0 ) {
            grid_r[i] = voxel_rpscace[i];
        }
    }

    for( int l = 0 ; l < nsurf ; l++ ) {
        const int lc = ng1 * surf_atom[l];

        for( int i = 0 ; i < ng1 ; i++ ) {
            if( xd[lc+i] > radius_surf2[surf_atom[l]] ) continue;
            const int ib = ng2 * i;

            for( int j = 0 ; j < ng1 ; j++ ) {
                if( yd[lc+j] > radius_surf2[surf_atom[l]] ) continue;
                const float d2 = xd[lc+i] + yd[lc+j];
                const int   jb = ib + ng1 * j;

                for( int k = 0 ; k < ng1 ; k++ ) {
                    const float d3 = d2 + zd[lc+k];

                    if( d3 <= radius_surf2[surf_atom[l]] ) {
                        grid_r[jb+k] = 3.0 * rho;
                    }
                }
            }
        }
    }

#endif // megadock 2.5.3 voxel

    delete[] voxel_rpscace;
    return;
}





//============================================================================//
void Receptor::electro(const float &beta,const float &eratio,
                       const int &num_grid,float *grid_coord,float *mol_coord,
                       int *iwork,float *fwork)
//============================================================================//
{
    const float   ftr1    = 6.0;          // FTDock parameter
    const float   ftr2    = 8.0;          // FTDock parameter
    const float   er1 = 4.0;          // FTDock parameter
    const float   er2 = 38.0;         // FTDock parameter
    const float   er3 = -224.0;       // FTDock parameter
    const float   er4 = 80.0;         // FTDock parameter

    float     *grid_r;
    float     *grid_i;
    const float   *xd, *yd, *zd;

    const int ng1 = num_grid;
    const int ng2 = ng1 * ng1;
    const int ng3 = ng1 * ng1 * ng1;
    const int na  = num_atoms();
    const int nag = num_atoms() * ng1;


    int search_range;
    float     er, d3;

    grid_r    = fwork;
    grid_i    = &fwork[ng3];
    xd        = &fwork[ng3*3+na*3];
    yd        = &fwork[ng3*3+na*3+nag];
    zd        = &fwork[ng3*3+na*3+nag*2];

#ifdef NEWVOXEL_ELEC

    const float grid_width = grid_coord[1] - grid_coord[0];
    const float charge_per_grid = -1 * grid_width * er4 / (beta * eratio);//formula derived from e(r) (r>8A)

    for( int l = 0 ; l < na ; l++ ) {
        const float abs_charge = fabs(_Charge[l]);
        if( abs_charge <= EPS ) continue;

        if (abs_charge < 0.07) { //hardcode (Gridsize = 1.2A)
            search_range = 4;
        } else if (abs_charge < 0.21) {
            search_range = 5;
        } else {
            search_range = (int)(abs_charge / charge_per_grid);
        }

        const int lc = ng1 * l;
        const int i2 = mol_coord[3*l  ] / grid_width + ng1 / 2; //limit search range
        const int j2 = mol_coord[3*l+1] / grid_width + ng1 / 2;
        const int k2 = mol_coord[3*l+2] / grid_width + ng1 / 2;
        const int ia = max(i2 - search_range, 0);
        const int ja = max(j2 - search_range, 0);
        const int ka = max(k2 - search_range, 0);
        const int ib = min(i2 + search_range+1, ng1);
        const int jb = min(j2 + search_range+1, ng1);
        const int kb = min(k2 + search_range+1, ng1);

        for( int i = ia ; i < ib ; i++ ) {// grid around atom[l]
            const int ijk_a = ng2 * i;

            for( int j = ja ; j < jb ; j++ ) {
                const int ijk_b = ijk_a + ng1 * j;
                const float d2 = xd[lc+i]+yd[lc+j];

                for( int k = ka ; k < kb ; k++ ) {
                    const int ijk = ijk_b + k;

                    if( grid_r[ijk] < 0.0 ) continue; //this grid is core of receptor

                    d3 = d2 + zd[lc+k];
                    d3 = sqrt(d3);

                    if( d3 >= ftr2 ) {
                        er = er4;
                    } else if( d3 > ftr1 ) {
                        er = er2*d3 + er3;
                    } else {
                        er = er1;
                    }
                    const float elec = eratio * beta * _Charge[l] / er / d3;
                    grid_i[ijk] += elec;

                }
            }
        }
    }


#else //NEWVOXEL_ELEC ####################################################
// megadock 2.5.3 voxel

    float d2;

    for( int l = 0 ; l < na ; l++ ) {
        if( fabs(_Charge[l]) <= EPS ) continue;
        const int lc = ng1 * l;

        for( int i = 0 ; i < ng1 ; i++ ) {
            for( int j = 0 ; j < ng1 ; j++ ) {
                d2 = xd[lc+i] + yd[lc+j];

                for( int k = 0 ; k < ng1 ; k++ ) {
                    const int kb = ng2*i + ng1*j + k;
                    if( grid_r[kb] < 0.0 ) continue;

                    d3 = d2 + zd[lc+k];
                    d3 = sqrt(d3);

                    if( d3 <= ftr1 ) {
                        er = er1;
                    } else if( d3 < ftr2 ) {
                        er = er2*d3 + er3;
                    } else {
                        er = er4;
                    }

                    grid_i[kb] += eratio * beta * _Charge[l] / er / d3;
                }
            }
        }
    }

#endif // megadock 2.5.3 voxel

    return;
}

//============================================================================//
void Receptor::precalc(const int &num_grid,int *nearesta,float *rjudge2,
                       float *radius_core2,float *radius_surf2)
//============================================================================//
{
    const float   D = 3.6;            // ZDOCK parameter
    const float   rcore2 = 1.5;           // ZDOCK parameter
    const float   rsurf2 = 0.8;           // ZDOCK parameter

    const int na  = num_atoms();
    const int ng1 = num_grid;
    const int ng3 = ng1 * ng1 * ng1;

    for( int i = 0 ; i < na ; i++ ) {
        rjudge2[i]      = (_Radius[i] + D)*(_Radius[i] + D);
        radius_core2[i] = _Radius[i] * _Radius[i] * rcore2;
        radius_surf2[i] = _Radius[i] * _Radius[i] * rsurf2;
    }

    for( int i = 0 ; i < ng3 ; i++ ) {
        nearesta[i] = -99;
    }

    return;
}


//============================================================================//
void Receptor::rpsc(const int &num_grid,float *grid_coord,float *mol_coord,
                    int *iwork,float *fwork,
                    const float &param_rec_core, const float &param_lig_core)
//============================================================================//
{
    const float   rho = -3.0;         // MEGADOCK parameter
    //    const float   epsilon = -27.0;            // MEGADOCK parameter
    const float epsilon = param_rec_core;     // Tuning Parameter
    const float open_space = -77.0;    // temporary score
    const float grid_width = 1.2;

    int       *nearesta;
    int       *surf_atom;
    int       *counter;
    float     *grid_r, *grid_i;
    float     *rjudge2, *radius_core2, *radius_surf2;
    float     *distance3;
    float     *xd, *yd, *zd;

    const int ng1 = num_grid;
    const int ng2 = ng1 * ng1;
    const int ng3 = ng1 * ng1 * ng1;
    const int na  = num_atoms();
    const int nag = num_atoms() * ng1;


    nearesta  = iwork;
    surf_atom = &iwork[ng3];
    counter   = &iwork[ng3+na];
    grid_r    = fwork;
    grid_i    = &fwork[ng3];
    rjudge2   = &fwork[ng3*2];
    radius_core2  = &fwork[ng3*2+na];
    radius_surf2  = &fwork[ng3*2+na*2];
    distance3     = &fwork[ng3*2+na*3];
    xd        = &fwork[ng3*3+na*3];
    yd        = &fwork[ng3*3+na*3+nag];
    zd        = &fwork[ng3*3+na*3+nag*2];

    precalc(ng1,nearesta,rjudge2,radius_core2,radius_surf2);
    voxel_init(ng1,grid_coord,mol_coord,grid_r,grid_i,distance3,xd,yd,zd);

// megadock 2.5.3 voxel

    for( int i = 0 ; i < ng3 ; i++ ) {
        counter[i] = 0;
    }

    int   i, j, k, lc, kb;
    float d2, d3;

    #pragma omp parallel for private(i,j,k,lc,d2,d3,kb)
    for( int l = 0 ; l < na ; l++ ) {
        lc = ng1 * l;

        for( i = 0 ; i < ng1 ; i++ ) {
            if( xd[lc+i] > rjudge2[l] ) continue;

            for( j = 0 ; j < ng1 ; j++ ) {
                if( yd[lc+j] > rjudge2[l] ) continue;
                d2 = xd[lc+i] + yd[lc+j];

                for( k = 0 ; k < ng1 ; k++ ) {
                    d3 = d2 + zd[lc+k];
                    if( d3 > rjudge2[l] ) continue;
                    kb = ng2*i + ng1*j + k;

                    #pragma omp critical
                    {
                        if( d3 < distance3[kb] ) {
                            distance3[kb] = d3;
                            nearesta[kb] = l;
                        }

                        counter[kb] ++;
                    }
                }
            }
        }
    }

    for( int i = 0 ; i < ng3 ; i++ ) {
        if( nearesta[i] == -99 ) continue;

        if( distance3[i] <= rjudge2[nearesta[i]] ) {
            if( distance3[i] <= radius_core2[nearesta[i]] ) {
                grid_r[i] = 9.0 * rho;
            } else {
                grid_r[i] = (float)counter[i];
            }
        }
    }

    int   nsurf = 0;

    for( int i = 1 ; i < ng1-1 ; i++ ) {
        const int ib = ng2 * i;

        for( int j = 1 ; j < ng1-1 ; j++ ) {
            const int jb = ib + ng1 * j;

            for( int k = 1 ; k < ng1-1 ; k++ ) {
                const int kb = jb + k;

                if( grid_r[kb] == 9.0*rho ) {
                    /*
                    if( grid_r[kb-ng2] == (float)counter[kb-ng2] ||
                        grid_r[kb+ng2] == (float)counter[kb+ng2] ||
                        grid_r[kb-ng1] == (float)counter[kb-ng1] ||
                        grid_r[kb+ng1] == (float)counter[kb+ng1] ||
                        grid_r[kb-1  ] == (float)counter[kb-1  ] ||
                        grid_r[kb+1  ] == (float)counter[kb+1  ] ) {
                    */
// Bug start
                    if( grid_r[kb-ng2] == 0.0 ||
                            grid_r[kb+ng2] == 0.0 ||
                            grid_r[kb-ng1] == 0.0 ||
                            grid_r[kb+ng1] == 0.0 ||
                            grid_r[kb-1  ] == 0.0 ||
                            grid_r[kb+1  ] == 0.0 ) {
// Bug end
                        if( distance3[kb] > radius_surf2[nearesta[kb]] ) {
                            grid_r[kb] = -9999.0;
                        } else {
                            int l;

                            for( l = 0 ; l < nsurf ; l++ ) {
                                if( surf_atom[l] == nearesta[kb] ) {
                                    break;
                                }
                            }

                            if( l >= nsurf ) {
                                surf_atom[nsurf++] = nearesta[kb];
                            }
                        }
                    }
                }
            }
        }
    }

    for( int i = ng2 ; i < ng3-ng2 ; i++ ) {
        if( grid_r[i] == -9999.0 ) {
            grid_r[i] = (float)counter[i];
        }
    }

    for( int l = 0 ; l < nsurf ; l++ ) {
        const int lc = ng1 * surf_atom[l];

        for( int i = 0 ; i < ng1 ; i++ ) {
            if( xd[lc+i] > radius_surf2[surf_atom[l]] ) continue;
            const int ib = ng2 * i;

            for( int j = 0 ; j < ng1 ; j++ ) {
                if( yd[lc+j] > radius_surf2[surf_atom[l]] ) continue;
                const float d2 = xd[lc+i] + yd[lc+j];
                const int   jb = ib + ng1 * j;

                for( int k = 0 ; k < ng1 ; k++ ) {
                    const float d3 = d2 + zd[lc+k];

                    if( d3 <= radius_surf2[surf_atom[l]] ) {
                        grid_r[jb+k] = 3.0 * rho;
                    }
                }
            }
        }
    }



    return;
}
